from typing import NamedTuple, Iterable, Optional

from runner.configs import get_cmd

RUN_NAME = 'run'


def set_run_name(name: str):
    global RUN_NAME
    RUN_NAME = name


def _get_task_name(config: dict) -> str:
    config_name = config.get('config_name')
    if config_name is not None:
        return RUN_NAME + ':' + config_name
    return RUN_NAME


class Task(NamedTuple):
    group_id: int
    cmd: str
    task_name: str
    memory_needed: float
    output: Optional[str]
    exit_code: Optional[int] = None


def set_exit_code(task: Task, code: int) -> Task:
    return Task(
        group_id=task.group_id,
        cmd=task.cmd,
        task_name=task.task_name,
        memory_needed=task.memory_needed,
        output=task.output,
        exit_code=code
    )


def _get_tasks(group_id: int, config_group: Iterable[dict]) -> Iterable[Task]:
    for config in config_group:
        yield Task(
            cmd=get_cmd(config),
            task_name=_get_task_name(config),
            memory_needed=config['memory_needed'],
            group_id=group_id,
            output=config.get('output')
        )



import asyncio
import logging
import os
import subprocess
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from itertools import chain
from pathlib import Path
from time import time, sleep
from typing import Iterable, Optional, Tuple

from nvitop import Device
from psutil import Process

from runner.task import Task, set_exit_code
from runner.util import to_gb, prepare_cmd

logger = logging.getLogger()
loop = ThreadPoolExecutor()


class GPUWatcher:

    def __init__(self, gpu: int, *, wait_time: float = 300, utilization_limit: float = 0.9):
        self._device = Device(gpu)
        self._reserved_space = 0.0
        self._occupied_space = 0.0
        self._foreign_pids = None
        self._own_pids = set()
        self._unavailable_until = 0.0
        self._wait_time = wait_time
        self._utilization_limit = utilization_limit
        self._finished_tasks = []
        self._last_task_id = 0

        self._update()

    def submit_task(self, task: Task):
        self._update()
        self._reserved_space += task.memory_needed
        loop.submit(self._run_task, task)

    def _run_task(self, task: Task) -> None:
        logger.info(f'Task is being put on device {self._device.index}!')

        redirect = asyncio.subprocess.PIPE
        if task.output is not None:
            output_path = Path(task.output)
            os.makedirs(output_path.parent, exist_ok=True)
            redirect = open(task.output, 'a')

        child_env = os.environ.copy()
        child_env['CUDA_VISIBLE_DEVICES'] = str(self._device.index)
        child_env['TASK_NAME'] = task.task_name
        child_process = subprocess.Popen(prepare_cmd(task.cmd), stdout=redirect, stderr=redirect, env=child_env)

        child_pid = child_process.pid
        self._own_pids.add(child_pid)
        child_process.wait()
        self._own_pids.remove(child_pid)

        self._reserved_space -= task.memory_needed

        finished_task = set_exit_code(task, child_process.returncode)
        self._finished_tasks.append(finished_task)

        if task.output is not None:
            redirect.close()

    def get_finished(self) -> Tuple[Task, ...]:
        finished = tuple(self._finished_tasks)
        self._finished_tasks = []
        return finished

    @property
    def available_memory(self) -> float:
        self._update()

        # give time for foreign processes to allocate memory
        if time() < self._unavailable_until:
            logger.debug(f'[device:{self._device.index}] Rejected. Reason: waiting on foreign process.')
            return 0.0

        # do not make available if utilization limit is exceeded
        if self._device.gpu_utilization() >= self._utilization_limit:
            logger.debug(f'[device:{self._device.index}] Rejected. Reason: GPU utilization is too high.')
            return 0.0

        return to_gb(self._device.memory_total()) - self._occupied_space - self._reserved_space

    @property
    def status(self) -> str:
        self._update()
        return f'Total: {to_gb(self._device.memory_total())}\n' \
               f'Occupied: {self._occupied_space}\n' \
               f'Reserved: {self._reserved_space}\n' \
               f'Available: {self.available_memory}\n' \
               f'Running PIDs: {self._own_pids}'

    def _update(self):
        processes = self._device.processes()

        # update own pids with their children
        for pid in self._own_pids:
            children_pids = set(child.pid for child in Process(pid).children(recursive=True))
            self._own_pids.update(children_pids)

        current_foreign_pids = {pid for pid, process in processes.items() if pid not in self._own_pids}
        if self._foreign_pids is None:
            self._foreign_pids = current_foreign_pids

        if set(current_foreign_pids).difference(set(self._foreign_pids)):
            # new processes appeared
            logger.info(f'Foreign processor detected on device {self._device.index}: '
                        f'fp{self._foreign_pids}, '
                        f'cfp{current_foreign_pids}, '
                        f'op{self._own_pids}')
            self._unavailable_until = time() + self._wait_time

        self._foreign_pids = current_foreign_pids
        bytes_memory = (processes[pid].gpu_memory() for pid in self._foreign_pids)
        self._occupied_space = sum(map(to_gb, bytes_memory))


class Scheduler:

    def __init__(
            self,
            gpus: Optional[Iterable[int]] = None,
            *,
            check_interval: float = 1.0,
            concurrent_jobs: Optional[int] = None,
            wait_time: float = 300,
            utilization_limit: float = 0.9
    ):
        if gpus is None:
            gpus = range(Device.count())
        watcher_builder = partial(GPUWatcher, wait_time=wait_time, utilization_limit=utilization_limit)
        self._watchers = tuple(map(watcher_builder, gpus))
        self._last_check_time = 0
        self._check_interval = check_interval
        self._concurrent_jobs = concurrent_jobs
        self._running_jobs = 0

    def schedule(self, task: Task) -> bool:
        # check for job limit
        if self._concurrent_jobs is not None and self._running_jobs >= self._concurrent_jobs:
            return False

        # check for time interval
        wait_for = self._check_interval - (time() - self._last_check_time)
        if wait_for > 0:
            sleep(wait_for)

        self._last_check_time = time()

        # find available gpu
        free_watcher = None
        for watcher in self._watchers:
            if watcher.available_memory >= task.memory_needed:
                free_watcher = watcher
                break

        if free_watcher is None:
            logger.debug(f'No GPU is available.')
            return False

        free_watcher.submit_task(task)
        self._running_jobs += 1
        return True

    def get_finished(self) -> Tuple[Task, ...]:
        finished = tuple(chain.from_iterable(map(GPUWatcher.get_finished, self._watchers)))
        self._running_jobs -= len(finished)
        return finished



from copy import deepcopy
from itertools import product
from typing import Iterable, Tuple, Union, List


def _collect_variable_args(args: Iterable[Union[str, List[str]]]) -> Tuple[Tuple[int, ...], Tuple[Tuple[str, ...], ...]]:
    """Collect positions and values of variable argument. Returns (positions, values)."""
    positions = []
    values = []
    for arg_idx, arg in enumerate(args):
        if isinstance(arg, list):
            positions.append(arg_idx)
            values.append(tuple(arg))

    return tuple(positions), tuple(values)


def _populate_arguments(values: Iterable[str], positions: Iterable[int], config: dict) -> None:
    if 'args' not in config:
        return
    args = config['args']
    for value, position in zip(values, positions):
        args[position] = value


def extract_configs(variable_configs: Iterable[dict]) -> Iterable[Tuple[dict, ...]]:
    """Returns configs grouped by origin"""
    for config in variable_configs:
        args = config.get('args', tuple())

        variable_positions, variable_values = _collect_variable_args(args)

        if not len(variable_values):
            yield deepcopy(config),
            continue

        new_configs = []

        for arg_variations in product(*variable_values):
            new_config = deepcopy(config)
            _populate_arguments(arg_variations, variable_positions, new_config)
            new_configs.append(new_config)

        yield tuple(new_configs)


def get_cmd(config: dict) -> str:
    """Get shell command from config without variables"""
    return ' '.join([config['command']] + config.get('args', []))



from setuptools import setup, find_packages
from io import open
from os import path

import pathlib

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# automatically captured required modules for install_requires in requirements.txt and as well as configure dependency links
with open(path.join(HERE, 'requirements.txt'), encoding='utf-8') as f:
    def remove_version(req: str):
        return req.split('==')[0]
    all_reqs = tuple(map(remove_version, f.read().split('\n')))

install_requires = [x.strip() for x in all_reqs if ('git+' not in x) and (not x.startswith('#')) and (not x.startswith('-'))]

dependency_links = [x.strip().replace('git+', '') for x in all_reqs if 'git+' not in x]

setup(
    name='triage-runner',
    description='Script running tool for optimizing GPU memory utilization.',
    version='0.3.2',
    packages=find_packages(),  # list of all packages
    install_requires=install_requires,
    python_requires='>=3.6',
    entry_points='''
        [console_scripts]
        triage=runner.__main__:main
    ''',
    author="Viktor Scherbakov",
    keyword="Script runner, ML/DL experiments, task runner, GPU job manager",
    long_description=README,
    long_description_content_type="text/markdown",
    license='MIT',
    url='https://github.com/ViktorooReps/triage',
    download_url='https://github.com/ViktorooReps/triage/archive/0.0.1.tar.gz',
    dependency_links=dependency_links,
    author_email='viktoroo.sch@gmail.com',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.11",
    ],
    include_package_data=True,
)



import collections
from typing import Iterator, Callable, TypeVar, Any, List


def consume(iterator: Iterator):
    collections.deque(iterator, maxlen=0)


_Any = TypeVar('_Any')


def apply(fn: Callable[[_Any], Any], iterator: Iterator[_Any]) -> None:
    consume(map(fn, iterator))


def to_gb(_bytes: int) -> float:
    return _bytes / 1024 / 1024 / 1024


def prepare_cmd(cmd: str) -> List[str]:
    return cmd.replace('  ', ' ').strip().split(' ')



import datetime
import json
import logging
import os
from typing import Iterable, TextIO, Optional

import click as click
import tqdm

import pkg_resources
from click_loglevel import LogLevel

from runner.concurrency import run_config_groups
from runner.task import set_run_name
from runner.configs import extract_configs
from runner.scheduler import Scheduler
from runner.tqdm_logger import setup_logging

setup_logging()
logger = logging.getLogger()

RUN_CONFIG_HELP = """
Running several configs: $ triage run_config1.json run_config2.json run_config3.json  

Patterns can be used for config discovery as well: $ triage run_config*.json   

More on pattern syntax can be found here: https://docs.python.org/3.10/library/pathlib.html#pathlib.Path.glob

Run configurations are stored in JSON format. The sample run configuration looks like this:

\b
{  
  "memory_needed": 10.0,
  "config_name": "sample_config",
  "output": "logs/sample_config_output.log"
  "command": "python3 train.py",
  "args": [
    "arg1",
    "--arg2",
    ["--seed=1", "--seed=2", "--seed=3"],
    "--arg3=3"
  ]   
}   

Every entry in `args` list is an argument for `command`. 
An entry can be a list - in which case TRIAGE will iterate 
through all the possible combinations of all values in list entries. 
The example script above will be run 3 times with an argument 
`--seed` set to 1, 2 and 3.

Parameter `config_name` is optional and is used for logging 
the results (see `--logfile` option). Based on this parameter 
environment variable `RUN_NAME` is set in order to be used by running script.

Parameter `output` is optional as well. Its value will be used as a path to
a file that will be used to redirect stdout and stdin of the running
command.
\f
"""

__version__ = pkg_resources.require('triage-runner')[0].version


def format_gpus(gpus: str) -> Iterable[int]:
    return map(int, gpus.split(','))


@click.command(help=RUN_CONFIG_HELP)
@click.argument('run_configs', nargs=-1, type=click.File('r'))
@click.option('-n', '--run_name', type=click.STRING,
              help='Run name. Will be used for setting RUN_NAME environment variable. Datetime by default')
@click.option('-v', '--version', is_flag=True,
              help='Print overseer version.')
@click.option('-j', '--jobs', type=click.INT,
              help='Maximum number of concurrently running configs. No limit by default.')
@click.option('-g', '--gpus', type=click.STRING,
              help='What GPUs to use. All available by default. Format: 0,2,3,1')
@click.option('-i', '--gpu_ping_interval', type=click.FLOAT, default=10.0,
              help='Interval in seconds between GPU status checks.')
@click.option('-w', '--wait', type=click.FLOAT, default=300,
              help='Waiting time in seconds after new foreign process appeared on GPU.')
@click.option('-u', '--utilization_limit', type=click.FLOAT, default=0.9,
              help='GPU utilization limit. Ranges from 0.0 to 1.0.')
@click.option('-l', '--log_level', type=LogLevel(), default=logging.WARNING,
              help='Logging level.')
def cli(
        run_configs: Iterable[TextIO],
        run_name: Optional[str],
        version: bool,
        jobs: int,
        gpus: Optional[str],
        gpu_ping_interval: float,
        wait: float,
        utilization_limit: float,
        log_level: str
):
    if version:
        print(f'TRIAGE {__version__}')
        return

    if run_name is None:
        run_name = datetime.datetime.now().strftime('[%d_%b_%H:%M]')
    set_run_name(run_name)

    if gpus is not None:
        gpus = format_gpus(gpus)

    logger.setLevel(log_level)

    file_names = [run_config.name for run_config in run_configs]
    logger.info(f'Found {len(file_names)} config files.')

    def read_and_close(file: TextIO) -> dict:
        content: dict = json.load(file)
        file.close()
        return content

    variable_configs = tuple(map(read_and_close, run_configs))
    grouped_configs = tuple(extract_configs(variable_configs))

    config_names = [config.get('config_name') for config in variable_configs]

    tqdm_descriptions = [(cn if cn is not None else fn) for cn, fn in zip(config_names, file_names)]
    progress_bars = [tqdm.tqdm(desc=desc, total=len(cf_group)) for desc, cf_group in zip(tqdm_descriptions, grouped_configs)]

    run_config_groups(
        grouped_configs,
        progress_bars,
        Scheduler(gpus, check_interval=gpu_ping_interval, concurrent_jobs=jobs, wait_time=wait, utilization_limit=utilization_limit)
    )


main = cli


if __name__ == '__main__':
    main()
